#ifndef ACC_CONF_H
#define ACC_CONF_H

/* ACC crate address */
#define ACC_CRATE      0

/* ACC station number */
#define ACC_N         22

/* output register station number, 
   which must produce a PULSE by A(0)F(17) (e.g. Kaizu 3470) */
#define ACC_NOPR      19

/* end of interrupt service routine (bit 9 counting from 0) */
#define ACC_ENDISR 0x200

/* maximum event length (in word) */
#define ACC_MAXEVL     250

/* crate address of Scalers */
#define SCL_CRATE      0

/* starting station number of Scalers */
#define SCL_N          1

/* "destructive read" function of Scalers */
#define SCL_F          2

/* Scaler DATA size (must be n times 12 or 16 depending on the module) */
#define SCL_NUM_DATA  16

/* Uncomment the following line to activate the double buffer, which 
 * will reduce the dead time by 25 msec/block, at most for 3922 C.C., 
 * but at the sacrifice of the stability of data acqusition.  
 * Do not use LATCH mode of Gate Generator for busy-veto. */
//#define ACC_DOUBLE_BUFFER

/*====== Followings must not be modified ? (if not power user...) ======*/

/* block size in 16-bit word */
#define ACC_BLK_SIZE  (16384/sizeof(short))

/* address pointer to scaler data in a 16-bit block data */
#define SCL_BLK_OFS  (ACC_BLK_SIZE-2*SCL_NUM_DATA)

#define ACC_BASE  0x00200400
/* Host => ACC instruction defining register */
#define ACC_IDR  (ACC_BASE+0x00)
/* ACC command response register */
#define ACC_RSR  (ACC_BASE+0x04)
/* ACC status register */
#define ACC_STR  (ACC_BASE+0x08)
/* ACC => Host instruction defining register */
#define ACC_VIR  (ACC_BASE+0x0C)
/* Host command response register */
#define ACC_VRR  (ACC_BASE+0x10)
/* Host status register */
#define ACC_VSR  (ACC_BASE+0x14)
/* Address of Buffer Prologue Table */
#define ACC_BPT  (ACC_BASE+0x18)
/* Address of Message Prologue Table */
#define ACC_MPT  (ACC_BASE+0x1C)

/* IDR: clear */
#define ACC_IDR_CLR  0x0001
/* IDR: change buffer */
#define ACC_IDR_XFR  0x0002
/* IDR: start DAQ */
#define ACC_IDR_STA  0x0004
/* IDR: stop DAQ */
#define ACC_IDR_STO  0x0008
/* IDR: abort DMA */
#define ACC_IDR_ABO  0x0010

/* RSR: error */
#define ACC_RSR_ERR  0x8000
/* RSR: done */
#define ACC_RSR_DONE 0x0001

/* STR: command done */
#define ACC_STR_DONE 0x0001
/* STR: DMA flag */
#define ACC_STR_DMA  0x0004
/* STR: buffer error */
#define ACC_STR_BERR 0x0040
/* STR: command error */
#define ACC_STR_CERR 0x0080
/* STR: end of interrupt1 */
#define ACC_STR_EISR 0x0200
/* all other STR bits used in j11.mac are not used */

/* VIR: buffer 1 full */
#define ACC_VIR_BUF1 0x0002
/* VIR: buffer 2 full */
#define ACC_VIR_BUF2 0x0004
/* VIR: buffer error */
#define ACC_VIR_ERR  0x0100

/* Address of BCB for current write */
#define ACC_BCBW(BPTP) (BPTP+0x00)
/* Address of BCB for current read */
#define ACC_BCBR(BPTP) (BPTP+0x08)
/* Address of BCB for data-full */
#define ACC_BCBF(BPTP) (BPTP+0x10)

/* header configuration */
#define ACC_HDR_LEN  4

/* block ID */
#define ACC_BLK_ID(b)  (b)[0]
/*   start comment block */
#define ACC_BID_STA  0x0001
/*   stop  comment block */
#define ACC_BID_STO  0xffff
/*   good raw data */
#define ACC_BID_RAW  0x0000
/*   bad data (serious error e.g. CAMAC overrun, DMA error, etc.) */
#define ACC_BID_BAD  0xff00

/* event count in this block */
#define ACC_BLK_EVC(b) (b)[1]

/* pointer to scaler address (in 16-bit array), but redundant ? */
#define ACC_SCL_ADR(b) (b)[2]/sizeof(short)

/* end of available data */
#define ACC_BLK_LEN(b) (b)[3]/sizeof(short)

/* Event header ID (MSB) */
#define ACC_EVT_ID     0xEE00

/* Event header */
#define ACC_EVT_HDR(b) (ACC_EVT_ID|(b&0x00ff))

#endif /* ACC_CONF_H */
