/*
 *  commander -- PHOBOS main control program
 */

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <sys/ioctl.h>
#include "camac.h"
#include "k3922_reg.h"
#include "acc_conf.h"
#include "crashm.h"

#define CMD_PATH        1
#define CMD_START       2
#define CMD_STOP        3
#define CMD_TRANSFER    4
#define CMD_BLOCKS      5
#define CMD_SCALER      6
#define CMD_RUNNO       7
#define CMD_RESET       8
#define CMD_EXIT       99

const struct {char *name; int min, key; }
cmd_table[] = {
  {"path"      ,  1, CMD_PATH      },
  {"start"     ,  3, CMD_START     },
  {"stop"      ,  3, CMD_STOP      },
  {"transfer"  ,  1, CMD_TRANSFER  },
  {"blocks"    ,  1, CMD_BLOCKS    },
  {"scaler"    ,  2, CMD_SCALER    },
  {"runno"     ,  2, CMD_RUNNO     },
  {"reset"     ,  2, CMD_RESET     },
  {"exit"      ,  1, CMD_EXIT      },
  {""          ,  0, 0             }
};

int chk_cmd(char *arg)
{
  char s[256]; int i, n;
  while (*arg == ' ') arg++;
  for (n = 0; *arg != ' ' && *arg != 0; arg++)
    s[n++] = ((*arg < 'a') ? *arg + ' ' : *arg); s[n] = 0;
  for (i = 0; cmd_table[i].key != 0; i++) {
    if (n >= cmd_table[i].min && n <= strlen(cmd_table[i].name))
      if (!strncmp(s, cmd_table[i].name, n)) return cmd_table[i].key;
  }
  return -1;
}

int rmt_path();
int chk_sampler();

struct  naf_data cc;
#define ccnaf(N,A,F,WSIZE) \
cc.mode = SINGLE | WSIZE;\
cc.naf  = NAFGEN(N,A,F);\
ioctl(fd, CCIOC_SINGLE, &cc);
#define ccnafw(N,A,F,WSIZE,D) \
cc.data = D;\
ccnaf(N,A,F,WSIZE);
#define ccnafq(N,A,F,WSIZE) \
cc.mode = SINGLE | WSIZE;\
cc.naf  = NAFGEN(N,A,F);\
{int i=0;\
do ioctl(fd, CCIOC_SINGLE, &cc);\
while ((cc.xq & (NOX | NOQ)) && (i++ < 10));}
#define ccnafqw(N,A,F,WSIZE,D) \
cc.data = D;\
ccnafq(N,A,F,WSIZE);

#define set_time(a,b)  {\
time_t tp=time(NULL);\
strftime(a, 9,"%H:%M:%S",localtime(&tp));\
strftime(b,10,"%d-%b-%y",localtime(&tp));}

int     shmid, semid;
CRASHM *shmp;

int     fd;

int acc_cmd(int cmd)
{
#define MAX_CMD_TRY 10
  int  i;
  ccnafqw(ACC_N,0,17,WSIZE24,ACC_IDR);  /* set ACC MAR to IDR */
  ccnafqw(ACC_N,0,16,WSIZE16,cmd);      /* write instruction to ACC */
  ccnafqw(ACC_N,0,17,WSIZE24,ACC_RSR);  /* set ACC MAR to RSR */
  ccnafqw(ACC_N,0,16,WSIZE16,0);        /* clear RSR */
  ccnafq (ACC_N,1,25,WSIZE24);  /* send interrupt to ACC (execute command) */
  for (i = 0; i < MAX_CMD_TRY; i++) {
    ccnafqw(ACC_N,0,17,WSIZE24,ACC_RSR); /* set ACC MAR to RSR */
    ccnafq (ACC_N,0, 0,WSIZE16);         /* read RSR */
    if (cc.data & (ACC_RSR_DONE | ACC_RSR_ERR)) break;
    usleep(10000L); /* wait a little (10 msec) */
  }
  return ((i == MAX_CMD_TRY) || (cc.data & ACC_RSR_ERR));
}

int main(int argc, char *argv[])
{
  int    i, v;
  char   s[256], path[256]="\0";
  struct sembuf semcmd;

  setvbuf(stdout, (char *)NULL, _IONBF, 0); /* set stdout unbuffered */

  if ((shmid = shmget(SHMKEY, sizeof(CRASHM), IPC_CREAT|0600)) == -1)
    {puts("! Cannot create shared memory"); exit(2);}
  if ((semid = semget(SEMKEY, SEM_NUM       , IPC_CREAT|0600)) == -1)
    {puts("! Cannot create semaphore"); exit(2);}
  if ((fd = open("/dev/cc0", O_RDWR)) == -1)
    {puts("! CAMAC device open error."); exit(2);}
  ioctl(fd, CCIOC_SET_CRATE, ACC_CRATE);
  /* check crate on-line */
  ccnaf(K3922_N,K3922_STAT_A,K3922_READ_F,WSIZE24);
  if (cc.data & K3922_STAT_OFL)
    {puts("! Crate off-line"); close(fd); exit(2);}

  shmp = (CRASHM *)shmat(shmid, 0, 0);
  if (!(argc > 1 && !strcmp(argv[1], "-n"))) {
    shmp->wp = shmp->runno = -1;
    shmp->acquire = shmp->remote = shmp->fd = shmp->nblock = shmp->mtape = 
      shmp->user[0] = shmp->host[0] = shmp->pwd[0] = shmp->path[0] = 0;
    for (i = 0; i < MAX_SAMPLERS; i++) shmp->sampler[i].pid = 0;}

  /* enable CC SRQ (which asserts Z to CAMAC bus !!) */
  ccnafw(K3922_N,K3922_STAT_A,K3922_WRIT_F,WSIZE24,K3922_STAT_SRQ);
  /* set LAM MASK to ACC */
  ccnafw(K3922_N,K3922_LAMM_A,K3922_WRIT_F,WSIZE24,(1<<(ACC_N-1)));

  while (gets(s)) {
    switch (chk_cmd(s)) {
    case CMD_PATH:
      if (shmp->acquire) puts("# Stop DAQ first"); else gets(path);
      break;
    case CMD_START:
      if (shmp->acquire) {puts("# Stop before start"); break;}
      gets(shmp->comment);
      shmp->remote = shmp->mtape = 0;
      if (path[0]) if (rmt_path(path)) break; path[0]=0;
      bzero((char *)shmp->scaler, sizeof(int)*SCL_NUM_DATA); /* clear scaler */
      chk_sampler();
      shmp->runno++;
      shmp->wp = shmp->nblock = shmp->badblock = 0;
      for (i = 0; i < MAX_SAMPLERS; i++) shmp->sampler[i].block = 0;
      set_time(shmp->sta_time, shmp->sta_date);
      shmp->sto_time[0] = shmp->sto_date[0] = 0;
      sem_op(semid, semcmd, sem_record, 1);
      ioctl(fd, CCIOC_WAKE_LAM, 0); /* notify recorder to write */
      sem_op(semid, semcmd, sem_record, 0); /* wait for xfer done */
      ccnafq(ACC_N,0,10,WSIZE24); /* clear ACC LAM */
      ccnafq(ACC_N,0,26,WSIZE24); /* enable ACC LAM */
      if (acc_cmd(ACC_IDR_CLR))   /* clear ACC buffer */
	{puts("# Error in clearing buffer, maybe broken ACC prog."); break;}
      if (acc_cmd(ACC_IDR_STA))   /* start ACC acquisition */
	{puts("# Error in staring DAQ, maybe broken ACC prog."); break;}
      ioctl(fd, CCIOC_ENABLE_INT, 0); /* enable interrupt from ACC */
      shmp->acquire = 1;
      printf("<< START >>\n%4.4d\n", shmp->runno);
      break;
    case CMD_STOP:
      if (!shmp->acquire) {puts("# Start before stop"); break;}
      sem_op(semid, semcmd, sem_record, 1);
      if (acc_cmd(ACC_IDR_STO)) /* stop DAQ: force last buf. xfered */
	puts("# Error in stopping DAQ, maybe broken ACC prog.");
      sem_op(semid, semcmd, sem_record, 0); /* wait for xfer done */
      ccnafq(ACC_N,0,24,WSIZE24); /* disable ACC LAM */
      ioctl(fd, CCIOC_DISABLE_INT, 0);
      shmp->acquire = 0;
      gets(shmp->comment);
      set_time(shmp->sto_time, shmp->sto_date);
      puts("<< STOP >>");
      usleep(50000L); /* this delay is required, but why ? */
      shmp->wp++;
      sem_op(semid, semcmd, sem_record, 1);
      ioctl(fd, CCIOC_WAKE_LAM, 0); /* notify recorder to write */
      sem_op(semid, semcmd, sem_record, 0); /* wait for xfer done */
      chk_sampler();
      break;
    case CMD_TRANSFER:
      if (!shmp->acquire) break;
      sem_op(semid, semcmd, sem_record, 1);
      acc_cmd(ACC_IDR_XFR);
      sem_op(semid, semcmd, sem_record, 0); /* wait for xfer done */
      break;
    case CMD_BLOCKS:
      chk_sampler();
      printf("<< BLOCKS >>\n%d blks", shmp->nblock);
      if (shmp->mtape) {
	v = shmp->mtblock*ACC_BLK_SIZE*sizeof(short);
	printf(" (%d.%d MB on tape)", v/1048576, (v%1048576)/104858);}
      puts("");
      break;
    case CMD_SCALER:
      chk_sampler();
      set_time(shmp->pri_time,shmp->pri_date);
      printf("RUN-%4.4d  START => %8s  STOP => %8s   Print => %8s  %9s\n",
	     shmp->runno, shmp->sta_time, shmp->sto_time, shmp->pri_time,
	     shmp->sta_date);
      printf("%s\n", shmp->comment);
      for (i = 0; i < SCL_NUM_DATA; i++) printf("%10d", shmp->scaler[i]);
      puts("");
      for (i = 0; i < MAX_SAMPLERS; i++)
	if (shmp->sampler[i].pid)
	  printf("%3d%6d: %-20s%10d (%10d)\n", i, shmp->sampler[i].pid,
		 shmp->sampler[i].name, shmp->sampler[i].block, shmp->nblock);
      break;
    case CMD_RUNNO:
      if (scanf("%d\n", &v) == 1 && v >= 0 && v < 10000)
	shmp->runno = v;
      else
	puts("# Invalid Run No.");
      break;
    case CMD_RESET:
      if (shmp->acquire) {puts("# Stop before reset"); break;}
      ioctl(fd, CCIOC_RESET, 0);
      break;
    case CMD_EXIT:
      if (shmp->acquire) {puts("# Stop before exit"); break;}
      goto EXIT;
    default:
//    printf("# Invalid Command (%s).\n", s);
      printf("Invalid Command.\n"); /* not to make passwd appear on screen...*/
      break;
    }
  }
 EXIT:
  shmp->acquire = -1;
  ioctl(fd, CCIOC_WAKE_LAM, 0); /* relieve recorder from WAIT_LAM loop */
  sem_op(semid, semcmd, sem_newdat, 1); /* wake up sampler(s) if waiting */
  /* delay is not necessary, because EIDRM error does the same job */
  usleep(100000L);
  close(fd);
  shmdt((char *)shmp);
  shmctl(shmid, IPC_RMID, 0);
  semctl(semid, 0, IPC_RMID, 0);
  exit(0);
}

#include <netdb.h>
#include <sys/socket.h>
#include <pwd.h>

char   sh[]="\
path=%s
file=run%4.4d.dat
if [ -c $path ]
  then
    if [ -w $path ]
      then {
        MT=`mt status -f $path` \
                             || { echo e\\# $path is not available.; exit 2; }
        echo $MT | grep -q ONLINE \
                             || { echo e\\# media not loaded.; exit 2; }
        echo $MT | grep -q WR_PROT \
                             && { echo e\\# write-protected media.; exit 2; }
        echo $MT | grep -q BOT \
                             && echo t0 || echo t
      }
      else { echo e\\# $path is write protected.; exit 2;}
    fi
  elif [ -d $path ]
    then
      if [ -w $path ]
        then
          if [ -e $path/$file ]
            then
              if [ -w $path/$file ]
                then { mv $path/$file $path/$file~; echo d }
                else echo e\\# $path/$file exists and write-protected.
              fi
            else echo d
          fi
        else echo e\\# $path is write protedted.
      fi
  else echo e\\# $path is invalid path
fi
";

struct servent *serv;

int rmt_path(char *path)
{
  char   *q, *user=(char *)NULL, *host=(char *)NULL, s[2048];
  int     sd=0;
  FILE   *fd;
  if ((shmp->remote = ((q = strchr(path, ':')) != NULL))) {
    host = path; *q = 0; path = ++q;
    if ((q = strchr(host, '@')) != NULL) {user = host; *q = 0; host = ++q;}
    else user = cuserid((char *)NULL);
    if (strcmp(host, shmp->host) || strcmp(user, shmp->user)) {
      if (isatty(0)) {       /* invoked from terminal */
	sprintf(s, "Password (%s:%s): ", host, user);
	strcpy(shmp->pwd, getpass(s));
      } else {               /* invoked as subprocess */
	printf("? Password (%s:%s):\n", host, user);
	gets(shmp->pwd);
      }
      strcpy(shmp->host, host); strcpy(shmp->user, user);
    }
  }
  strcpy(shmp->path, path);
  sprintf(s, sh, path, shmp->runno+1);
  if (!shmp->remote) {
    if ((fd = popen(s, "r+")) == NULL) {puts("# popen error"); return -1;}
  } else {
    if ((serv = getservbyname("exec", "tcp")) == NULL)
      {puts("# exec/tcp: service not available"); return -1;}
    if ((sd = rexec((char **)(&host), serv->s_port, user,
		    shmp->pwd, "/bin/sh", (int *)NULL)) == -1)
      {puts("# rexec fail"); shmp->host[0] = 0; return -1;}
    fd = fdopen(sd, "r+");
    fputs(s, fd); fflush(fd);
  }
  s[0] = 0;
  if (fgets(s, 256, fd) == NULL) puts("# rexec script error");
  if (!shmp->remote) pclose(fd);
  else {fclose(fd); shutdown(sd, 2); close(sd);}
  switch (s[0]) {
  case 't':
    shmp->mtape = 1;
    printf("<< OPEN >>\n%s (tape", path);
    if (s[1] == '0') {printf(", new volume"); shmp->mtblock =0;}
    if (shmp->remote) printf(") on %s@%s\n", user, host);
    else puts(")");
    break;
  case 'd':
    shmp->mtape = 0;
    printf("<< OPEN >>\n%s/run%4.4d.dat", path, shmp->runno+1);
    if (shmp->remote) printf(" on %s@%s", user, host);
    puts("");
    break;
  case 'e':
    puts((char *)&s[1]);
    return -1;
  default:
    printf("# unexpected error (%s)\n", s);
    return -1;
  }
  return 0;
}

int chk_sampler()
{
  FILE *fd;
  int   p[1024], i, j, n=0;
  char  s[100];
  fd = popen("ps xch", "r");
  while (fgets(s, 100, fd) != NULL)
    if ((p[n] = atoi(s))) {if (++n >= 1024) return 1;}
  pclose(fd);
  for (i = 0; i < MAX_SAMPLERS; i++) {
    for (j = 0; j < n; j++) if (shmp->sampler[i].pid == p[j]) goto found;
    shmp->sampler[i].pid = 0;
  found:
  }
  return 0;
}
